/**
 * EnhProj.e -- $Id: EnhProj.e#8 2009/06/28 12:04:33 chrisant $
 *
 *    Enhanced project utilities.  Initially this is the Slick-C portion of an
 *    enhanced file open dialog, which integrates with a matching EnhProj.dll.
 *    Later this may gain additional abilities.
 *
 * Features:
 *
 *    - The file list is automatically filtered as you type.
 *    - Use the tabs to list Workspace files, Project files, or Any files from
 *      the file system (i.e. starting from the current working directory).
 *
 *    - Keyboard shortcuts in the Open dialog:
 *       - Ctrl+M = Smart tab.
 *       - Ctrl+F = Any tab.
 *       - Ctrl+P = Project tab.
 *       - Ctrl+W = Workspace tab.
 *       - Ctrl+H = History tab.
 *       - Ctrl+O = Open file with the standard open dialog.
 *       - Ctrl+N or Shift+Enter or Ctrl+Enter = Create new file.
 *
 * Configuration:
 *
 *    Look for "enhproj" below for a list of configuration variables that
 *    control optional behavior.  At this time they can only be set in the
 *    editor via "set-var enhproj_variable_name" at the command line.  At some
 *    point in the future I may hook up a way to set them from the UI.
 *
 * How to filter the list:
 *
 *    - Type a filename prefix to filter the list.
 *    - Type a path\filename prefix to filter by path as well as filename.
 *    - Use * and ? wildcards to filter path or filename using wildcards.
 *    - Type "*foo" to filter the list to files containing "foo" somewhere in
 *      their name.
 *    - Examples:
 *          - Type "foo" to see files that begin with "foo".
 *          - Type "*foo" to see files with "foo" in their name.
 *          - Type "bar\foo" to see files that begin with "foo" in the "bar"
 *            directory.
 *          - Type "bar*\*foo" to see files with "foo" in their name in any
 *            directory that begins with "bar".
 */
/*
 * TODO:
 *
 *    - Maybe merge the History list into the Smart list?
 *    - Hook up UI for choosing options.
 */

#include "slick.sh"
#pragma option( strict, on )
#region Imports
#import "menu.e"
#endregion

#ifndef FILEHIST_CATEGORY
// SE14 defines this in menu.e, but earlier versions do not have it.
#define FILEHIST_CATEGORY 'filehist'
#endif

// ??? -- SE14 seems to get confused and refuse to compile this without at
// least one extern function prototype declared.
extern SomeRandomNonExistentFunctionNameToPacifyCompiler();

enum { MODE_SMART, MODE_FILESYSTEM, MODE_PROJECT, MODE_WORKSPACE, MODE_HISTORY };
enum { LIST_FILESYSTEM, LIST_PROJECT, LIST_WORKSPACE, LIST_HISTORY };

// These are the default settings.  Use 'set-var enhproj_ofd_which_list' or
// etc to configure them inside the editor.
int enhproj_ofd_which_list = MODE_SMART;  // -1 = remember across editor sessions, else a MODE_xyz value.
int enhproj_ofd_remembered_list = MODE_SMART; // a MODE_xyz value
int enhproj_ofd_sort_by_path = 0;         // 0 = sort by name, 1 = sort by path
int enhproj_ofd_case_sensitive = -1;      // 0 = case insensitive, 1 = case sensitive
int enhproj_ofd_always_substring = 1;     // 0 = match file prefix and path suffix, 1 = match file substring and path suffix, 2 = match file substring and path substring
int enhproj_ofd_smart_includes_workspace = 1; // 0 = smart merges LIST_FILESYSTEM + LIST_PROJECT, 1 = smart merges LIST_FILESYSTEM + LIST_WORKSPACE

int enhproj_ofd_x = 0;                    // saved x position of dialog
int enhproj_ofd_y = 0;                    // saved y position of dialog
int enhproj_ofd_cx = 0;                   // saved width of dialog
int enhproj_ofd_cy = 0;                   // saved height of dialog

#ifdef CHRISANT
int enhproj_dbg_dll = 0;                  // 0 = use ship dll, 1 = use debug dll
int enhproj_unload_always = 0;            // 0 = unload at process exit, 1 = unload after closing dialog
#endif

static _str s_fDllLoaded = false;

definit()
{
   if ( arg( 1 ) == '' )
      s_fDllLoaded = false;
   if ( enhproj_ofd_case_sensitive < 0 )
      enhproj_ofd_case_sensitive = ( int )( _fpos_case == '' );
}

/**
 * <ul>
 *    <li>1 = GetProjectFiles(), <b>~740ms</b>
 *    <li>2 = absolute(), <b>~340ms</b>
 *    <li>3 = chdir() and GetFullPathName(), <b>~360ms</b>
 *    <li>4 = vsFileAbsolute(), <b>~440ms</b>
 *    <li>5 = optimized native code to fully qualify path, <b>???</b>
 * </ul>
 */
#define RELPATHS_STYLE 5   //$ WARNING: (chrisant) MUST MATCH PROJUTIL.H!!!

static void AppendFilesFromCurrentBuffer( int which_list, _str working_dir )
{
   _str filename;

#if RELPATHS_STYLE == 3
   _str oldCwd = getcwd();
   chdir( working_dir, 1 );
#elif RELPATHS_STYLE == 5
   FlatFile_SetRelativeDir( working_dir );
#endif

   FlatFile_AllocateFileList( which_list, FlatFile_Count( which_list ) + p_Noflines );
   top();
   up();
   while ( !down() )
   {
      get_line( filename );
      FlatFile_AppendFile( which_list,
#if RELPATHS_STYLE == 1 || RELPATHS_STYLE == 3 || RELPATHS_STYLE == 4 || RELPATHS_STYLE == 5
                           filename
#elif RELPATHS_STYLE == 2
                           absolute( filename, working_dir )
#endif
                           );
   }

#if RELPATHS_STYLE == 3
   chdir( oldCwd, 1 );
#elif RELPATHS_STYLE == 5
   FlatFile_SetRelativeDir( '' );
#endif
}

void EnhProj_RefreshHistoryFiles()
{
   int file_view_id = 0;
   int orig_view_id = _create_temp_view( file_view_id );

   if ( !_menu_find(_mdi.p_menu_handle, FILEHIST_CATEGORY, auto dash_mh, auto dash_pos, 'C') )
   {
      int numItems = _menu_info( dash_mh, 'C' );
      for ( ii := dash_pos + 1; ii < numItems; ++ii )
      {
         _menu_get_state( dash_mh, ii, auto flags, 'p', auto caption, auto command );
         parse command with . command;
         command = strip( stranslate( command, '', '"' ), 'T' );
         file_view_id.insert_line( command );
      }
   }

   FlatFile_ClearFiles( LIST_HISTORY );
   AppendFilesFromCurrentBuffer( LIST_HISTORY, '' );
   FlatFile_SortFiles( LIST_HISTORY );

   activate_window( orig_view_id );
   _delete_temp_view( file_view_id );
}

void EnhProj_RefreshProjectFiles()
{
   if ( _project_name != '' )
   {
      if ( FlatFile_IsRefreshNeeded( LIST_PROJECT, _project_name ) )
      {
         int file_view_id = 0;
         int orig_view_id = _create_temp_view( file_view_id );

         GetProjectFiles( _project_name, file_view_id,
                          '', null, '', false,
                          RELPATHS_STYLE == 1, // true = absolute paths (may be slow), false = relative to project directory
                          false, // not listbox format
                          -1, '' );

         FlatFile_ClearFiles( LIST_PROJECT );
         FlatFile_AddAssociatedFile( LIST_PROJECT, _project_name );
         AppendFilesFromCurrentBuffer( LIST_PROJECT, strip_filename( _project_name, 'N' ) );
         FlatFile_SortFiles( LIST_PROJECT );

         activate_window( orig_view_id );
         _delete_temp_view( file_view_id );
      }
   }
   else
   {
      FlatFile_ClearFiles( LIST_PROJECT );
   }
}

void EnhProj_RefreshWorkspaceFiles()
{
   if ( _workspace_filename != '' )
   {
      int ii;
      _str projects[] = null;
      _str mostRecentProjectFile = '';
      boolean fRefresh = !FlatFile_Count( LIST_WORKSPACE );

      _GetWorkspaceFiles( _workspace_filename, projects );

      if ( FlatFile_IsRefreshNeeded( LIST_WORKSPACE, _workspace_filename ) )
         fRefresh = true;
      for ( ii = 0; ii < projects._length(); ii++ )
      {
         projects[ii] = _AbsoluteToWorkspace( strip_filename( projects[ii], 'E' ) :+ PRJ_FILE_EXT );
         if ( FlatFile_IsRefreshNeeded( LIST_WORKSPACE, projects[ii] ) )
            fRefresh = true;
      }

      if ( fRefresh )
      {
         FlatFile_ClearFiles( LIST_WORKSPACE );

//#define TIME_COLLECT_FILES
#ifdef TIME_COLLECT_FILES
typeless timeBeginCollectWorkspaceFiles = _time( 'B' );
#endif

         FlatFile_AddAssociatedFile( LIST_WORKSPACE, _workspace_filename );
         for ( ii = 0; ii < projects._length(); ii++ )
         {
            int file_view_id = 0;
            int orig_view_id = _create_temp_view( file_view_id );

            projects[ii] = _AbsoluteToWorkspace( strip_filename( projects[ii], 'E' ) :+ PRJ_FILE_EXT );
            GetProjectFiles( projects[ii], file_view_id,
                             '', null, '', false,
                             RELPATHS_STYLE == 1, // true = absolute paths (may be slow), false = relative to project directory
                             false, // not listbox format
                             -1, '' );

            FlatFile_AddAssociatedFile( LIST_WORKSPACE, projects[ii] );
            AppendFilesFromCurrentBuffer( LIST_WORKSPACE, strip_filename( projects[ii], 'N' ) );

            activate_window( orig_view_id );
            _delete_temp_view( file_view_id );
         }

#ifdef TIME_COLLECT_FILES
typeless timeEndCollectWorkspaceFiles = _time( 'B' );
say( 'TOOK '( timeEndCollectWorkspaceFiles - timeBeginCollectWorkspaceFiles )' MILLISECONDS TO COLLECT WORKSPACE FILES' );
#endif

         FlatFile_SortFiles( LIST_WORKSPACE );
      }
   }
   else
   {
      FlatFile_ClearFiles( LIST_WORKSPACE );
   }
}

_command void enhOpen() name_info(','VSARG2_READ_ONLY|VSARG2_MARK)
{
   if ( !s_fDllLoaded )
   {
      int status = -1;
      _str s;

#ifdef CHRISANT
      // Construct the path to where a locally built copy of the dll would be.
      s = 'c:\chrisant\projutil\bin.ntx86\';
      if ( enhproj_dbg_dll )
         s :+= 'dbg';
      else
         s :+= 'rtl';
      s :+= '\enhproj.dll';
      status = _dllload( s );
#endif

      if ( status )
      {
         s = slick_path_search( 'enhproj.dll' );
         if ( s._length() )
            status = _dllload( s );
      }

      if ( !status )
         s_fDllLoaded = true;
   }

   int mode = enhproj_ofd_which_list;
   if ( mode < 0 )
      mode = enhproj_ofd_remembered_list;

   FlatFile_SetOptions( enhproj_ofd_sort_by_path,
                        enhproj_ofd_case_sensitive,
                        enhproj_ofd_always_substring,
                        enhproj_ofd_smart_includes_workspace );

   FlatFile_SetMode( mode );
   EnhProjDll_SetOfdPos( enhproj_ofd_x, enhproj_ofd_y, enhproj_ofd_cx, enhproj_ofd_cy );
   EnhProjDll_ShowOpenFileDialog();

   int newMode = FlatFile_GetMode();
   if ( enhproj_ofd_remembered_list != newMode ||
        EnhProjDll_IsOfdMoved() )
   {
      enhproj_ofd_remembered_list = newMode;
      enhproj_ofd_x = EnhProjDll_GetOfdX();
      enhproj_ofd_y = EnhProjDll_GetOfdY();
      enhproj_ofd_cx = EnhProjDll_GetOfdCX();
      enhproj_ofd_cy = EnhProjDll_GetOfdCY();
      // Let the editor know that it needs to save the state.
      _config_modify_flags( CFGMODIFY_DEFVAR );
   }

#ifdef CHRISANT
   if ( enhproj_unload_always )
      enhUnload();
#endif
}

_command void enhUnload() name_info(','VSARG2_READ_ONLY|VSARG2_MARK)
{
   if ( s_fDllLoaded )
   {
      /*
       * There is a nasty loader lock deadlock when trying to unload the
       * EnhProj.dll earlier than at process exit time.  The problem is that
       * vsDllExit is the natural point for EnhProj.dll to end its background
       * thread(s), and since the DLL is about to be unloaded it needs to wait
       * for the thread(s) to exit before returning from vsDllExit otherwise
       * the background thread(s) can crash when the DLL is unloaded out from
       * under the running code.  But it can't wait inside vsDllExit because
       * it is inside the loader lock, and the background thread(s) will
       * acquire the loader lock while exiting (see callstack below).  Thus it
       * sets up a deadly embrace.
       *
       * The workaround is for the macro code to be careful about how it
       * unloads the DLL:  the macro explicitly tells the DLL to clean up
       * before unloading it, so that the main thread is not holding the
       * loader lock while the DLL's background thread(s) exit.
       *
       * >  ntdll.dll!_KiFastSystemCallRet@0()
       *    ntdll.dll!_ZwWaitForSingleObject@12()  + 0xc bytes
       *    ntdll.dll!_RtlpWaitOnCriticalSection@8()  + 0xc8 bytes
       *    ntdll.dll!_RtlEnterCriticalSection@4()  - 0x4b84 bytes
       *    ntdll.dll!_LdrGetProcedureAddressEx@20()  + 0xc1 bytes
       *    ntdll.dll!_LdrGetProcedureAddress@16()  + 0x18 bytes
       *    kernel32.dll!_GetProcAddress@8()  + 0x3f bytes
       *    shell32.dll!7650e79d()
       *    [Frames below may be incorrect and/or missing, no symbols loaded for shell32.dll]
       *    kernel32.dll!_HeapFree@12()  + 0x14 bytes
       *    shell32.dll!7650e829()
       *    shell32.dll!76510f1f()
       *    ole32.dll!NotifyInitializeSpies()  + 0x51 bytes
       *    ole32.dll!_CoUninitialize@0()  + 0x36 bytes
       *    ole32.dll!DoThreadSpecificCleanup()  + 0x2edda bytes
       *    ole32.dll!FlsThreadCleanupCallback()  + 0x19 bytes
       *    ntdll.dll!_RtlProcessFlsData@4()  + 0x787 bytes
       *    ntdll.dll!_LdrShutdownThread@0()  + 0x35 bytes
       *    ntdll.dll!_RtlExitUserThread@4()  + 0x2a bytes
       *    msvcrt.dll!__endthreadex()  + 0x23 bytes
       *    msvcrt.dll!__endthreadex()  - 0x25 bytes
       *    msvcrt.dll!__endthreadex()  + 0xf2 bytes
       *    msvcrt.dll!__endthreadex()  + 0xf2 bytes
       *    kernel32.dll!@BaseThreadInitThunk@12()  + 0x12 bytes
       *    ntdll.dll!___RtlUserThreadStart@8()  + 0x27 bytes
       *    ntdll.dll!__RtlUserThreadStart@8()  + 0x1b bytes
       */
      EnhProjDll_StopThreads();
      dunload( "enhproj.dll" );
      s_fDllLoaded = false;
   }
}
