#include <stdlib.h>
#include "arrays.h"

/*
 * RgBase - Base class for array template.
 */

RgBase::RgBase( unsigned const cbItem )
	: m_prg( 0 )
	, m_cItems( 0 )
	, m_cAlloc( 0 )
	, m_cbItem( cbItem )
{
	Assert( m_cbItem );
}

RgBase::~RgBase()
{
	// Derived classes must call Free in the destructor for the proper
	// DestructElms to be called.

	Assert( !Count() );
	Assert( !m_prg );
}

bool
RgBase::SetSize( unsigned const cItems )
{
	// When growing, append elements to the end of the array.

	if( cItems > Count() )
		return BaseInsert( Count(), cItems - Count() );

	// When shrinking, remove elements from the end of the array.

	if( cItems < Count() )
		Remove( cItems, Count() - cItems );

	return true;
}

void
RgBase::Clear()
{
	Remove( 0, Count() );
}

void
RgBase::Free()
{
	Clear();

	free( m_prg );
	m_prg = 0;
	m_cAlloc = 0;

	Assert( !m_cItems );
}

RgBase::ELM*
RgBase::Item( unsigned const index ) const
{
	Assert( index <= Count() );
	Assert( m_cbItem );

	return m_prg + ( index * m_cbItem );
}

RgBase::ELM*
RgBase::Items() const
{
	return Count() ? m_prg : 0;
}

/*
 * RgBase::FSearch - Finds the matching occurrence with the lowest index.
 *
 * Note that 'pFind' is always passed as the first parameter to 'pfn'.
 *
 * Returns true if an exact match was found; *pIndex is the index of the
 * match.  Returns false if no match was found; *pIndex is the index at which
 * the element should be inserted.
 */

bool
FBinarySearch( void const* const pvFind, void* const pvCookie, unsigned const cItems, unsigned* pindex, BSearchFnPtr pfn )
{
	unsigned lo = 0;
	unsigned hi = cItems;
	unsigned mid = 0;
	bool fMatch = false;

	while( hi > lo )
	{
		mid = ( hi + lo ) / 2;

		int const result = pfn( pvFind, mid, pvCookie );

		if( result > 0 )
		{
			lo = mid + 1;
		}
		else
		{
			if( result == 0 )
			{
				fMatch = true;
				// Keep searching, to find first occurrence (ensures O(ln(n))
				// performance no matter what).
			}

			hi = mid;
		}
	}

	if( pindex )
		*pindex = lo;

	return fMatch;
}

bool
RgBase::FSearch( CompareFn pfn, void const* const pFind, unsigned* const pIndex, unsigned iLo, unsigned cItems ) const
{
	int n = 1;			// Any non-zero value will do.

	// Binary search.

	if( -1 == iLo )
	{
		iLo = 0;
		cItems = Count();
	}

	Assert( -1 != iLo );
	Assert( -1 != cItems );

	unsigned iHi = iLo + cItems;

	while( iLo < iHi )
	{
		unsigned iMid = ( iHi + iLo ) / 2;

		n = pfn( pFind, reinterpret_cast<void const*>( Item( iMid ) ) );

		if( n <= 0 )
			iHi = iMid;
		else
			iLo = iMid + 1;
	}

	if( pIndex )
		*pIndex = iLo;

	//$ review: (chrisant) Should be able to optimize this away, but be
	// careful not to break the case where we find the match early and
	// continue to loop.

	if( iLo < cItems )
		n = pfn( pFind, reinterpret_cast<void const*>( Item( iLo ) ) );
	else
		n = 1;

	return !n;
}

void
RgBase::Sort( CompareFn pfn )
{
	qsort( Items(), Count(), m_cbItem, pfn );
}

bool
RgBase::BaseInsert( unsigned index, unsigned cInsert, ELM** ppOut )
{
	Assert( 0 <= int( index ) && index <= Count() );

	if( cInsert <= 0 )
		return true;

	// Grow the array as necessary.

	if( !EnsureAllocation( Count() + cInsert ) )
		return false;

	// Shift items to make room for the new items.

	unsigned const cbMove = ( Count() - index ) * m_cbItem;
	ELM* const pInsertAt = Item( index );

	if( cbMove > 0 )
	{
		ELM* const pMoveTo = Item( index + cInsert );
		memmove( pMoveTo, pInsertAt, cbMove );
	}

	m_cItems += cInsert;

	// Run the constructors.

	ConstructElms( cInsert, pInsertAt );

	if( ppOut )
		*ppOut = ( cInsert == 1 ) ? pInsertAt : 0;

	return true;
}

void
RgBase::Remove( unsigned const index, unsigned const cRemove )
{
	Assert( 0 <= int( index ) && index < Count() || !cRemove );
	Assert( index + cRemove <= Count() );

	if( cRemove <= 0 )
		return;

	// Run the destructors.

	ELM* const pRemoveAt = Item( index );

	DestructElms( cRemove, pRemoveAt );

	// Shift items to close the gap in the array.

	unsigned const cbMove = ( Count() - ( index + cRemove ) ) * m_cbItem;

	if( cbMove > 0 )
	{
		ELM* const pMoveFrom = Item( index + cRemove );
		memmove( pRemoveAt, pMoveFrom, cbMove );
	}

	m_cItems -= cRemove;
}

void
RgBase::Compact( unsigned cItemsExtra )
{
	// Determine allocation size.

	unsigned const cAlloc = Min( Count() + cItemsExtra, m_cAlloc );

	// Free the memory entirely when possible.

	if( !cAlloc )
	{
		Free();
		return;
	}

	// Reallocate array.

	ELM* const pCompacted = static_cast<ELM*>( realloc( m_prg, cAlloc * m_cbItem ) );

	if( !pCompacted )
		return;

	m_prg = pCompacted;
	m_cAlloc = cAlloc;
}

bool
RgBase::EnsureAllocation( unsigned cItems )
{
	if( cItems <= m_cAlloc )
		return true;

	// Ensure O(n) performance when appending repeatedly.

	enum { c_cInitial = 8 };

	unsigned cAlloc = Max( ( m_cAlloc ? m_cItems << 1 : c_cInitial ), cItems );

	// Allocate array.

	Assert( FImplies( !m_cAlloc, !m_cItems ) );
	Assert( FImplies( !m_cAlloc, !m_prg ) );
	Assert( FImplies( m_cAlloc, m_prg ) );

	ELM* const pAlloc = reinterpret_cast<ELM*>( realloc( m_prg, cAlloc * m_cbItem ) );

	if( !pAlloc )
		return false;

	m_prg = pAlloc;
	m_cAlloc = cAlloc;

	Assert( m_prg );
	Assert( m_cAlloc >= cItems );

	return true;
}

void
RgBase::Swap( RgBase& other )
{
	Assert( m_cbItem == other.m_cbItem );

	ELM* const prg = m_prg;
	m_prg = other.m_prg;
	other.m_prg = prg;

	unsigned const cItems = m_cItems;
	m_cItems = other.m_cItems;
	other.m_cItems = cItems;

	unsigned const cAlloc = m_cAlloc;
	m_cAlloc = other.m_cAlloc;
	other.m_cAlloc = cAlloc;
}

