#pragma once

#include <new.h>
#include "std.h"

typedef int ( *CompareFn )( void const* pElm1, void const* pElm2 );

/*
 * RgBase - Base class for array template.
 */

class RgBase
{
public:
	unsigned		Count() const { return m_cItems; }

	bool			EnsureAllocation( unsigned cItems );

	bool			SetSize( unsigned cItems );
	void			Compact( unsigned cItemsExtra = 0 );
	void			Clear();
	void			Free();

	void			Remove( unsigned index, unsigned cRemove = 1 );
	void			Move( unsigned indexFrom, unsigned indexTo );

	bool			FSearch( CompareFn pfn, void const* pFind, unsigned* pIndex, unsigned iLo = -1, unsigned cItems = -1 ) const;
	void			Sort( CompareFn pfn );

protected:
	typedef unsigned char ELM;

					RgBase( unsigned cbItem );
	virtual			~RgBase();

	bool			BaseInsert( unsigned index, unsigned cInsert, ELM** ppOut = 0 );

	ELM*			Item( unsigned index ) const;
	ELM*			Items() const;

	virtual void	ConstructElms( unsigned cItems, void* prgItems ) const = 0;
	virtual void	DestructElms( unsigned cItems, void* prgItems ) const = 0;

private:
	void			Swap( RgBase& other );

					RgBase( RgBase const& );	// Forbidden.
	RgBase&			operator=( RgBase const& );	// Forbidden.

#ifdef DEBUG
protected:
#else
private:
#endif
	ELM*			m_prg;

private:
	unsigned		m_cItems;
	unsigned		m_cAlloc;
	unsigned const	m_cbItem;
};

/*
 * Rg - Array template.
 */

template <class T>
class Rg : public RgBase
{
	typedef RgBase base;

public:
					Rg();
	virtual			~Rg();

	T&				operator[]( int index ) const { return operator[]( unsigned( index ) ); }
	T&				operator[]( unsigned index ) const;

	T*				Item( unsigned index ) const;
	T*				Items() const { return reinterpret_cast<T*>( base::Items() ); }

					// ItemMax() is the pointer immediately after the last
					// item, useful for loop termination conditions.

	T*				ItemMax() const { return reinterpret_cast<T*>( base::Item( Count() ) ); }

	bool			Insert( unsigned index, T** ppOut = 0 );
	bool			Insert( unsigned index, unsigned cItems );
	bool			Append( T** ppOut = 0 ) { return Insert( Count(), ppOut ); }

	void			Swap( Rg<T>& other );

protected:
	void			ConstructElms( unsigned cItems, void* prgItems ) const OVERRIDE;
	void			DestructElms( unsigned cItems, void* prgItems ) const OVERRIDE;

#ifdef DEBUG
private:
	T**				m_pprgT;	// Get a better view of the array in the debugger.
#endif
};

template <class T>
Rg<T>::Rg()
	: RgBase( sizeof( T ) )
#ifdef DEBUG
	, m_pprgT( reinterpret_cast<T**>( &m_prg ) )
#endif
{
}

template <class T>
Rg<T>::~Rg()
{
	// Must call this here so that the correct DestructElms is called.

	Free();
}

template <class T>
T&
Rg<T>::operator []( unsigned index ) const
{
	Assert( index < Count() );

	return *Item( index );
}

template <class T>
T*
Rg<T>::Item( unsigned index ) const
{
	if( index < Count() )
		return reinterpret_cast<T*>( base::Item( index ) );

	Assert( !index && !Count() );

	return 0;
}

template <class T>
bool
Rg<T>::Insert( unsigned index, T** ppOut )
{
	return BaseInsert( index, 1, reinterpret_cast<ELM**>( ppOut ) );
}

template <class T>
bool
Rg<T>::Insert( unsigned index, unsigned cItems )
{
	return BaseInsert( index, cItems );
}

template <class T>
void
Rg<T>::Swap( Rg<T>& other )
{
	RgBase::Swap( other );

#ifdef DEBUG
	T** const ppT = m_pprgT;
	m_pprgT = other.m_pprgT;
	other.m_pprgT = ppT;
#endif
}

template <class T>
void
Rg<T>::ConstructElms( unsigned cItems, void* const prgItems ) const
{
	ZeroMemory( prgItems, cItems * sizeof( T ) );

	for( T* prgTyped = reinterpret_cast<T*>( prgItems ); cItems--; ++prgTyped )
		::new( prgTyped ) T;
}

template <class T>
void
Rg<T>::DestructElms( unsigned cItems, void* const prgItems ) const
{
#ifdef DEBUG
	unsigned const cbFill = cItems * sizeof( T );
#endif

	for( T* prgTyped = reinterpret_cast<T*>( prgItems ); cItems--; ++prgTyped )
		prgTyped->~T();

#ifdef DEBUG
	memset( prgItems, 0xe5, cbFill );
#endif
}

/*
 * FBinarySearch - simple generic binary search.
 */

typedef int ( *BSearchFnPtr )( void const* pvFind, unsigned indexCompare, void* pvCookie );

bool FBinarySearch( const void *pvFind, void* pvCookie, unsigned cItems, unsigned short* pindex, BSearchFnPtr pfn );

