#include "projutil.h"
#include <vsapi.h>
#include "std.h"
#include "str.h"
#include "arrays.h"
#include "flatfile.h"
#include "icons.h"
#include "res.h"

struct AssocFileForRefresh
{
	Str m_sFile;
	FILETIME m_ftModified;
};

static int s_iMode = MODE_SMART;
static Rg<FileInfo>* s_rgprgFiles[LIST_MAX] = {};
static Rg<FileInfo*>* s_prgMatches = 0;
static Rg<AssocFileForRefresh>* s_rgprgAssocFiles[LIST_MAX] = {};

#if RELPATHS_STYLE == 4 || RELPATHS_STYLE == 5
static Str s_sRelativeDrive;
static Str s_sRelativeDir;
#endif

static bool s_fSortByPath = false;
static int s_nAlwaysSubstringMatch = 0;
bool g_fSmartIncludesWorkspace = true;

#if 0
inline ULONGLONG
FileTimeInMS( FILETIME const* const pft )
{
	return *reinterpret_cast<ULONGLONG const*>( pft ) / 10000i64;
}
#endif

static bool
GetFileTimeByName( char const* const pszFile, FILETIME* const pft )
{
	ZeroMemory( pft, sizeof( *pft ) );

	HANDLE const h = CreateFile( pszFile, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, 0, OPEN_EXISTING, 0, 0 );
	if( h == INVALID_HANDLE_VALUE )
		return false;

	bool const fSuccess = !!GetFileTime( h, 0, 0, pft );

	DWORD const dwError = GetLastError();
	CloseHandle( h );
	SetLastError( dwError );

	return fSuccess;
}

/*
 * FileInfo.
 */

FileInfo::FileInfo()
	: m_iImage( II_NONE )
	, m_fEnsuredIcon( false )
{
}

bool
FileInfo::EnsureSysIcon( HWND hwndNotify )
{
	if( !m_fEnsuredIcon )
	{
		Assert( !m_pAsyncShellIcon );

		m_fEnsuredIcon = true;
		m_pAsyncShellIcon = new AsyncShellIcon();

		if( m_pAsyncShellIcon )
		{
			Str s;
			s.Set( m_sPath );
			s.Append( m_sFile );
			m_pAsyncShellIcon->Enqueue( hwndNotify, s.Text(), SHGFI_SMALLICON );
		}
	}

	if( m_pAsyncShellIcon && m_pAsyncShellIcon->IsComplete() )
	{
		m_iImage = m_pAsyncShellIcon->RetrieveIconIndex();
		m_pAsyncShellIcon->Release();
		m_pAsyncShellIcon = 0;
	}

	return m_iImage != II_NONE;
}

/*
 * Probe the OS compare/sort functions to determine the order in which they
 * sort characters.  This is used by our custom compare/sort function that
 * matches full strings, prefix strings, or wildcarded strings.
 *
 * Caveat:  The order will be wrong for multibyte characters.  Sorry, but
 * that's significantly more complex to get right, and I have to limit the
 * time I invest on this.
 */

class CharAsString
{
public:
	CharAsString( char const* psz ) { m_sz[0] = *psz; m_sz[1] = 0; }
	operator char const*() { return m_sz; }
private:
	char m_sz[2];
};

static int
cmp_CharAsString( void const* pv1, void const* pv2 )
{
	CharAsString s1( static_cast<char const*>( pv1 ) );
	CharAsString s2( static_cast<char const*>( pv2 ) );
	return lstrcmp( s1, s2 );
}

static int
cmpi_CharAsString( void const* pv1, void const* pv2 )
{
	CharAsString s1( static_cast<char const*>( pv1 ) );
	CharAsString s2( static_cast<char const*>( pv2 ) );
	return lstrcmpi( s1, s2 );
}

static int
cmpiSort_CharAsString( void const* pv1, void const* pv2 )
{
	CharAsString s1( static_cast<char const*>( pv1 ) );
	CharAsString s2( static_cast<char const*>( pv2 ) );
	int const n = lstrcmpi( s1, s2 );
	if( n )
	    return n;
	return lstrcmp( s1, s2 );
}

class CharsInOrder
{
	enum { ORDER_STRICT, ORDER_CASELESS };
public:
	CharsInOrder();
	unsigned char const* GetOrder() const { return m_rgCharsInOrder[ORDER_STRICT]; }
	unsigned char const* GetCaselessOrder() const { return m_rgCharsInOrder[ORDER_CASELESS]; }
private:
	unsigned char m_rgCharsInOrder[2][256];
};

CharsInOrder::CharsInOrder()
{
	unsigned char rg[256];

	for( unsigned iStyle = 2; iStyle--; )
	{
		unsigned char* const prgCharsInOrder = m_rgCharsInOrder[iStyle];

		for( unsigned ii = 0; ii < DimensionOf( rg ); ii++ )
			rg[ii] = ii;

		if( iStyle == ORDER_STRICT )
			qsort( rg + 1, DimensionOf( rg ) - 3, sizeof( rg[0] ), cmp_CharAsString );
		else
			qsort( rg + 1, DimensionOf( rg ) - 3, sizeof( rg[0] ), cmpiSort_CharAsString );

		unsigned iEquivalent = 0;
		for( unsigned ii = 0; ii < DimensionOf( rg ); ii++ )
		{
			if( iStyle == ORDER_STRICT )
				iEquivalent = ii;
			else if( cmpi_CharAsString( &rg[iEquivalent], &rg[ii] ) )
				iEquivalent = ii;

			prgCharsInOrder[rg[ii]] = iEquivalent;
		}
	}
}

static CharsInOrder s_CharsInOrder;

/*
 * Case insensitive sort helpers.
 */

enum CmpSortFlags
{
	CSF_NONE				= 0x00,
	CSF_WILDCARDS			= 0x01,
	CSF_CASEINSENSITIVE		= 0x02,
	CSF_STARSTOPSATSLASH	= 0x04,
	CSF_SORT				= 0x08,
};
inline CmpSortFlags operator|( CmpSortFlags const& a, CmpSortFlags const& b ) { return CmpSortFlags( int( a ) | int( b ) ); }
inline CmpSortFlags operator&( CmpSortFlags const& a, CmpSortFlags const& b ) { return CmpSortFlags( int( a ) & int( b ) ); }

static CmpSortFlags s_flagsDefault = CSF_CASEINSENSITIVE;

static int
cmpsort( char const* psz1, int _cch1, char const* psz2, int _cch2, CmpSortFlags flags )
{
	//$ todo: (chrisant) Performance could be improved by just detecting null
	// terminators as we compare, but that would make it harder to support the
	// strnicmp usage pattern.
	if( _cch1 < 0 )
		_cch1 = lstrlen( psz1 ) + 1;
	if( _cch2 < 0 )
		_cch2 = lstrlen( psz2 ) + 1;

	int n = 0;
	int nStrict = 0;

	Trace_( "[%s]/%d vs [%s]/%d == ", psz1, _cch1, psz2, _cch2 );

	if( _cch1 || _cch2 )
	{
		CASSERT( static_cast<unsigned char>( char( 0xff ) ) == 0xff );

		char const* const pszEnd1 = psz1 + _cch1;
		char const* const pszEnd2 = psz2 + _cch2;

		// First match fixed prefix.

		while( psz1 < pszEnd1 )
		{
			char const ch2Raw = psz2 < pszEnd2 ? *psz2 : 0;

			if( flags & CSF_WILDCARDS && *psz1 == '*' )
				break;

			if( flags & CSF_CASEINSENSITIVE )
			{
				unsigned char ch1 = s_CharsInOrder.GetCaselessOrder()[static_cast<unsigned char>( *psz1 )];
				unsigned char ch2 = s_CharsInOrder.GetCaselessOrder()[static_cast<unsigned char>( ch2Raw )];
				n = ( static_cast<int>( ch1 ) - static_cast<int>( ch2 ) );

				if( !nStrict )
				{
					ch1 = s_CharsInOrder.GetOrder()[static_cast<unsigned char>( *psz1 )];
					ch2 = s_CharsInOrder.GetOrder()[static_cast<unsigned char>( ch2Raw )];
					nStrict = ( static_cast<int>( ch1 ) - static_cast<int>( ch2 ) );
				}
			}
			else
			{
				unsigned char ch1 = s_CharsInOrder.GetOrder()[static_cast<unsigned char>( *psz1 )];
				unsigned char ch2 = s_CharsInOrder.GetOrder()[static_cast<unsigned char>( ch2Raw )];
				n = ( static_cast<int>( ch1 ) - static_cast<int>( ch2 ) );
			}

			if( flags & CSF_WILDCARDS && *psz1 == '?' && ch2Raw )
				n = 0;

			psz1++;
			if( psz2 < pszEnd2 )
				psz2++;

			if( n )
				break;
		}

		// Now handle star wildcards.

		if( flags & CSF_WILDCARDS && !n && psz1 < pszEnd1 && *psz1 == '*' )
		{
			struct Param
			{
				size_t		iStart;
				size_t		iEnd;
			};
			struct Params
			{
				char const*	buffer;
				Param		rgParam[5];
			};
			struct BackupStruct
			{
				char const*	pch;
				Param*		pParam;
			};

			Params params;
			params.buffer = psz2;

			BackupStruct rgBackups[5];
			BackupStruct* pNextBackup = rgBackups;
			BackupStruct* pBackup;

			for( ;; )
			{
				pBackup = pNextBackup;

				// Check for end of wildcard expression string.

				if( psz1 >= pszEnd1 )
				{
					Trace( "0  (Has wildcards)" );
					return 0;

LRetry:
					for( ;; --pBackup )
					{
						if( pBackup <= rgBackups )
						{
							Trace( "-1  (Failed to match, has wildcards)" );
							return -1;
						}

						psz1 = pBackup[-1].pch;
						psz2 = params.buffer + --( pBackup[-1].pParam->iEnd );

						Assert( psz1 < pszEnd1 );

						if( *psz1 != '*' && *psz1 != '?' )
						{
							Assert( psz2 < pszEnd2 );

							// Optimization:  If *psz1 != *psz2 we'll get here
							// again through a circuitous path just to
							// decrement pBackup[-1].pParam->iEnd.  So the
							// loop below backtracks more efficiently.

							char const* const pchStart = params.buffer + pBackup[-1].pParam->iStart;
							if( flags & CSF_CASEINSENSITIVE )
							{
								while( psz2 >= pchStart )
								{
									unsigned char const ch1 = s_CharsInOrder.GetCaselessOrder()[static_cast<unsigned char>( *psz1 )];
									unsigned char const ch2 = s_CharsInOrder.GetCaselessOrder()[static_cast<unsigned char>( *psz2 )];
									if( ch1 == ch2 )
										break;
									psz2--;
								}
							}
							else
							{
								while( psz2 >= pchStart )
								{
									if( *psz1 == *psz2 )
										break;
									psz2--;
								}
							}
							pBackup[-1].pParam->iEnd = psz2 - params.buffer;
						}

						if( psz2 >= params.buffer + pBackup[-1].pParam->iStart )
						{
							pNextBackup = pBackup;
							break;
						}
					}
					continue;
				}

				// Match characters.

				if( *psz1 == '*' )
				{
					pNextBackup++;
					if( pNextBackup - rgBackups > DimensionOf( rgBackups ) )
					{
						// Too many wildcards; give up and fail.
						Trace( "1  (Too many wildcards!)" );
						return 1;
					}

					pBackup->pParam = params.rgParam + ( pBackup - rgBackups );
					pBackup->pParam->iStart = psz2 - params.buffer;

					while( psz2 < pszEnd2 && ( !( flags & CSF_STARSTOPSATSLASH ) || *psz2 != '\\' ) )
						psz2++;

					while( psz1 < pszEnd1 && *psz1 == '*' )
						psz1++;

					pBackup->pParam->iEnd = psz2 - params.buffer;
					pBackup->pch = psz1;
				}
				else
				{
					while( psz1 < pszEnd1 )
					{
						if( *psz1 == '*' )
							break;

						if( psz2 >= pszEnd2 )
							goto LRetry;

						if( flags & CSF_WILDCARDS )
						{
							unsigned char ch1 = s_CharsInOrder.GetCaselessOrder()[static_cast<unsigned char>( *psz1 )];
							unsigned char ch2 = s_CharsInOrder.GetCaselessOrder()[static_cast<unsigned char>( *psz2 )];
							n = ( static_cast<int>( ch1 ) - static_cast<int>( ch2 ) );

							if( !nStrict )
							{
								ch1 = s_CharsInOrder.GetOrder()[static_cast<unsigned char>( *psz1 )];
								ch2 = s_CharsInOrder.GetOrder()[static_cast<unsigned char>( *psz2 )];
								nStrict = ( static_cast<int>( ch1 ) - static_cast<int>( ch2 ) );
							}
						}
						else
						{
							unsigned char ch1 = s_CharsInOrder.GetOrder()[static_cast<unsigned char>( *psz1 )];
							unsigned char ch2 = s_CharsInOrder.GetOrder()[static_cast<unsigned char>( *psz2 )];
							n = ( static_cast<int>( ch1 ) - static_cast<int>( ch2 ) );
						}

						if( flags & CSF_WILDCARDS && *psz1 == '?' && *psz2 )
							n = 0;

						psz1++;
						psz2++;

						if( n )
							goto LRetry;
					}
				}
			}
		}
	}

	if( flags & CSF_SORT )
	{
		if( !n )
			n = nStrict;
	}

	Trace( "%d", n );
	return n;
}

static int
Cmp_NoCase( char const* psz1, char const* psz2 )
{
	return cmpsort( psz1, -1, psz2, -1, s_flagsDefault );
}

static int
CmpSort_NoCase( char const* psz1, char const* psz2 )
{
	return cmpsort( psz1, -1, psz2, -1, s_flagsDefault|CSF_SORT );
}

static int
Cmp_WildNoCase( char const* psz1, char const* psz2 )
{
	return cmpsort( psz1, -1, psz2, -1, s_flagsDefault|CSF_WILDCARDS );
}

static int
Cmp_Lookup( char const* pszPartialFile, FileInfo const* pfi )
{
	char const* const psz1 = pszPartialFile;
	char const* const psz2 = pfi->m_sFile.Text();
	// Note that the null terminators are excluded from the comparison.
	return cmpsort( psz1, lstrlen( psz1 ), psz2, lstrlen( psz2 ), s_flagsDefault|CSF_WILDCARDS );
}

static int
Cmp_NonWildPrefix( char const* pszPartialFile, FileInfo const* pfi )
{
	char const* const psz1 = pszPartialFile;
	char const* const psz2 = pfi->m_sFile.Text();
	char const* const pszWild = strpbrk( pszPartialFile, "*?" );
	Assert( pszWild );
	// Note that the null terminators are excluded from the comparison.
	return cmpsort( psz1, pszWild - psz1, psz2, lstrlen( psz2 ), s_flagsDefault|CSF_WILDCARDS );
}

static int
Cmp_SortByName( void const* pv1, void const* pv2 )
{
	FileInfo const* const pfi1 = static_cast<FileInfo const*>( pv1 );
	FileInfo const* const pfi2 = static_cast<FileInfo const*>( pv2 );
	int n;
	// Sort directories ahead of files.
	n = int( pfi2->m_attr & FILE_ATTRIBUTE_DIRECTORY ) - int( pfi1->m_attr & FILE_ATTRIBUTE_DIRECTORY );
	if( n )
		return n;
	// Compare file names (caseless).
	n = Cmp_NoCase( pfi1->m_sFile.Text(), pfi2->m_sFile.Text() );
	if( n )
		return n;
	// If file names are same (caseless), use a sort-compare on paths.
	n = CmpSort_NoCase( pfi1->m_sPath.Text(), pfi2->m_sPath.Text() );
	if( n )
		return n;
	// If paths sort the same, use a sort-compare on file names in case the
	// entire path+file strings differed only by case.
	n = CmpSort_NoCase( pfi1->m_sFile.Text(), pfi2->m_sFile.Text() );
	if( n )
		return n;
	return 0;
}

static int
Cmp_SortByPath( void const* pv1, void const* pv2 )
{
	FileInfo const* const pfi1 = static_cast<FileInfo const*>( pv1 );
	FileInfo const* const pfi2 = static_cast<FileInfo const*>( pv2 );
	int n;
	// Sort directories ahead of files.
	n = int( pfi2->m_attr & FILE_ATTRIBUTE_DIRECTORY ) - int( pfi1->m_attr & FILE_ATTRIBUTE_DIRECTORY );
	if( n )
		return n;
	// Compare paths (caseless).
	n = Cmp_NoCase( pfi1->m_sPath.Text(), pfi2->m_sPath.Text() );
	if( n )
		return n;
	// If paths are same (caseless), use a sort-compare on file names.
	n = CmpSort_NoCase( pfi1->m_sFile.Text(), pfi2->m_sFile.Text() );
	if( n )
		return n;
	// If file names sort the same, use a sort-compare on paths in case the
	// entire path+file strings differed only by case.
	n = CmpSort_NoCase( pfi1->m_sPath.Text(), pfi2->m_sPath.Text() );
	if( n )
		return n;
	return 0;
}

/*
 * Helper functions.
 */

char const*
GetFileName( char const* pszFile )
{
	char const* pszName = pszFile;

	for( char const* pszWalk = pszFile; *pszWalk; pszWalk++ )
		if( '/' == *pszWalk || '\\' == *pszWalk )
			pszName = pszWalk + 1;

	return pszName;
}

char const*
GetExtension( char const* pszFile )
{
	char const* pszDot;

	for( pszDot = 0; *pszFile; pszFile++ )
		switch( *pszFile )
		{
		case '.':
			pszDot = pszFile;
			break;
		case ' ':
		case ':':
		case '\\':
		case '/':
			pszDot = 0;
			break;
		}

	return pszDot;
}

/*
 * File list miscellaneous public functions.
 */

extern "C" void VSAPI
FlatFile_Uninitialize()
{
	delete s_prgMatches;
	s_prgMatches = 0;

	for( unsigned ii = DimensionOf( s_rgprgFiles ); ii--; )
	{
		delete s_rgprgFiles[ii];
		s_rgprgFiles[ii] = 0;
	}

	for( unsigned ii = DimensionOf( s_rgprgAssocFiles ); ii--; )
	{
		delete s_rgprgAssocFiles[ii];
		s_rgprgAssocFiles[ii] = 0;
	}

#if RELPATHS_STYLE == 4 || RELPATHS_STYLE == 5
	s_sRelativeDrive.Free();
	s_sRelativeDir.Free();
#endif
}

extern "C" void VSAPI
FlatFile_SetOptions( int sort_by_path,
					 int case_sensitive,
					 int always_substring_match,
					 int smart_includes_workspace )
{
	s_fSortByPath = !!sort_by_path;
	s_flagsDefault = ( !case_sensitive ? CSF_CASEINSENSITIVE : CSF_NONE );
	s_nAlwaysSubstringMatch = always_substring_match;
	g_fSmartIncludesWorkspace = !!smart_includes_workspace;

	if( s_nAlwaysSubstringMatch < 0 || s_nAlwaysSubstringMatch > 2 )
		s_nAlwaysSubstringMatch = 0;
}

extern "C" void VSAPI
FlatFile_SetMode( ModeType iMode )
{
	delete s_prgMatches;
	s_prgMatches = 0;

	s_iMode = iMode;
}

extern "C" int VSAPI
FlatFile_GetMode()
{
	return s_iMode;
}

/*
 * Multiple lists of files, which get stitched together by the filtering
 * layer.
 */

extern "C" void VSAPI
FlatFile_ClearFiles( ListType iList )
{
	delete s_prgMatches;
	s_prgMatches = 0;

	for( int ii = DimensionOf( s_rgprgFiles ); ii--; )
	{
		if( iList == LIST_ALLLISTS || ii == iList )
		{
			delete s_rgprgFiles[ii];
			s_rgprgFiles[ii] = 0;
		}
	}

	for( int ii = DimensionOf( s_rgprgAssocFiles ); ii--; )
	{
		if( iList == LIST_ALLLISTS || ii == iList )
		{
			delete s_rgprgAssocFiles[ii];
			s_rgprgAssocFiles[ii] = 0;
		}
	}
}

extern "C" int VSAPI
FlatFile_IsRefreshNeeded( ListType iList, VSHVAR hvarFilename )
{
	if( iList < 0 || iList >= LIST_MAX )
		return false;

	char const* const pszFile = vsHvarGetZ( hvarFilename );
	if( !pszFile )
		return false;

	if( !s_rgprgAssocFiles[iList] )
		return true;

	FILETIME ft;
	GetFileTimeByName( pszFile, &ft );

	for( unsigned ii = s_rgprgAssocFiles[iList]->Count(); ii--; )
	{
		AssocFileForRefresh const* const p = s_rgprgAssocFiles[iList]->Item( ii );
		if( !Cmp_NoCase( pszFile, p->m_sFile.Text() ) )
			return ( CompareFileTime( &ft, &p->m_ftModified ) > 0 );
	}

	return true;
}

extern "C" int VSAPI
FlatFile_AddAssociatedFile( ListType iList, VSHVAR hvarFilename )
{
	if( iList < 0 || iList >= LIST_MAX )
		return -1;

	if( !s_rgprgAssocFiles[iList] )
	{
		s_rgprgAssocFiles[iList] = new Rg<AssocFileForRefresh>;
		if( !s_rgprgAssocFiles[iList] )
			return -1;
	}

	char const* const pszFile = vsHvarGetZ( hvarFilename );
	if( !pszFile )
		return -1;

	AssocFileForRefresh* pAssocFile;
	if( !s_rgprgAssocFiles[iList]->Append( &pAssocFile ) )
		return -1;

	pAssocFile->m_sFile.Set( pszFile );
	GetFileTimeByName( pszFile, &pAssocFile->m_ftModified );
	return 0;
}

extern "C" int VSAPI
FlatFile_AllocateFileList( ListType iList, unsigned cFiles )
{
	if( iList < 0 || iList >= LIST_MAX )
		return -1;

	if( !s_rgprgFiles[iList] )
	{
		s_rgprgFiles[iList] = new Rg<FileInfo>;
		if( !s_rgprgFiles[iList] )
			return -1;
	}

	if( !s_rgprgFiles[iList]->EnsureAllocation( cFiles ) )
		return -1;

	return 0;
}

extern "C" void VSAPI
FlatFile_SetRelativeDir( VSHVAR hvarRelativeDir )
{
#if RELPATHS_STYLE == 4 || RELPATHS_STYLE == 5
	char const* pszPath = vsHvarGetZ( hvarRelativeDir );
	if( pszPath && *pszPath )
	{
		Assert( pszPath[0] != '.' );

		s_sRelativeDrive.Set( pszPath );
		char const* const pszColon = strchr( s_sRelativeDrive.Text(), ':' );
		Assert( pszColon );	// NYI: UNC...
		s_sRelativeDrive.SetEnd( pszColon + 1 );

		s_sRelativeDir.Set( pszPath );
		if( s_sRelativeDir.Length() && s_sRelativeDir.Text()[s_sRelativeDir.Length() - 1] != '\\' )
			s_sRelativeDir.Append( '\\' );
	}
	else
	{
		s_sRelativeDrive.Clear();
		s_sRelativeDir.Clear();
	}
#endif
}

extern "C" int VSAPI
FlatFile_AppendFile( ListType iList, VSHVAR hvarFilename )
{
	if( iList < 0 || iList >= LIST_MAX || !s_rgprgFiles[iList] )
		return -1;

	char const* pszPath = vsHvarGetZ( hvarFilename );
	if( !pszPath )
		return -1;

	FileInfo* p;
	if( !s_rgprgFiles[iList]->Append( &p ) )
		return -1;

#if RELPATHS_STYLE == 3
	char szFullPath[MAX_PATH];
	char* pszFilePart;
	DWORD const cch = GetFullPathName( pszPath, DimensionOf( szFullPath ), szFullPath, &pszFilePart );
	if( cch && cch < DimensionOf( szFullPath ) )
		pszPath = szFullPath;
#elif RELPATHS_STYLE == 4
	char szFullPath[MAX_PATH];
	vsFileAbsoluteTo( szFullPath, pszPath, s_sRelativeDir.Text() );
	pszPath = szFullPath;
#elif RELPATHS_STYLE == 5
	// Paths are expected to be of one of the following forms:
	//	1.  c:\foo\bar\abc.h (nothing to do)
	//	2.  \\server\foo\bar\abc.h (nothing to do)
	//	3.  \foo\bar\abc.h (copy drive from s_sRelativeDir)
	//	4.  bar\abc.h (prepend s_sRelativeDir)
	//	5.  ..\..\..\foo\bar\abc.h (starting at s_sRelativeDir, go up a parent for each ..\, append what's left over)
	char szFullPath[MAX_PATH];
	if( ( pszPath[0] && pszPath[1] == ':' && pszPath[2] == '\\' ) ||
		( pszPath[0] == '\\' && pszPath[1] == '\\' ) ||
		iList == LIST_HISTORY )
	{
		// Path is already fully qualified.
	}
	else if( pszPath[0] == '\\' )
	{
		// Copy drive from s_sRelativeDir.
		sprintf_s( RgC( szFullPath ), "%s%s", s_sRelativeDrive.Text(), pszPath );
		pszPath = szFullPath;
	}
	else
	{
		Assert( !strchr( pszPath, ':' ) );
		Assert( pszPath[0] != '\\' );

		// Consume "..\" entries or ".\" entries.
		char const* const pszParent = s_sRelativeDir.Text();
		size_t cchParent = s_sRelativeDir.Length();
		while( pszPath[0] == '.' )
		{
			if( pszPath[1] == '.' && pszPath[2] == '\\' )
			{
				// Handle "..\" but assume that SlickEdit has given us a valid
				// path.  If it hasn't, we won't crash, but we may compute a
				// bogus result.  It's a trade-off to improve performance.

				if( cchParent > 0 )
				{
					// Strip trailing slash.
					if( pszParent[cchParent - 1] == '\\' )
						cchParent--;
					// Strip directory name.
					//$ WARNING: This may have trouble with DBCS encodings.
					char const* pszStrip = pszParent + cchParent;
					while( cchParent > 0 && *( pszStrip - 1 ) != '\\' )
					{
						cchParent--;
						pszStrip--;
					}
				}

				pszPath += 3;
			}
			else if( pszPath[1] == '\\' )
			{
				pszPath += 2;
			}
			else
			{
				break;
			}
		}

		// Prepend s_sRelativeDir.
		sprintf_s( RgC( szFullPath ), "%.*s%s", cchParent, s_sRelativeDir.Text(), pszPath );
		pszPath = szFullPath;
	}
#endif

	char const* const pszFile = GetFileName( pszPath );
	size_t const cchPath = ( pszFile - pszPath );

	p->m_sFile.Set( pszFile );
	p->m_sPath.Set( pszPath, cchPath );
	p->m_attr = 0;
#if 0
	ZeroMemory( &p->m_ft, sizeof( p->m_ft ) );
	ZeroMemory( &p->m_uliFileSize, sizeof( p->m_uliFileSize ) );
#endif

	return 0;
}

int
FlatFile_AppendFileData( ListType iList, char const* pszDir, WIN32_FIND_DATA const* pfd )
{
	Assert( iList == LIST_FILESYSTEM );

	if( !s_rgprgFiles[iList] )
		return -1;

	FileInfo* p;
	if( !s_rgprgFiles[iList]->Append( &p ) )
		return -1;

	Assert( *pszDir && *pfd->cFileName );
	p->m_fFromFileSystem = true;
	p->m_sFile.Set( pfd->cFileName );
	p->m_sPath.Set( pszDir );
	p->m_attr = pfd->dwFileAttributes;
	p->m_ft = pfd->ftLastWriteTime;
	p->m_uliFileSize.LowPart = pfd->nFileSizeLow;
	p->m_uliFileSize.HighPart = pfd->nFileSizeHigh;

	if( pfd->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
		p->m_sFile.Append( "\\" );

	return 0;
}

extern "C" void VSAPI
FlatFile_SortFiles( ListType iList )
{
	if( iList >= 0 && iList < LIST_MAX && s_rgprgFiles[iList] )
	{
		if( iList == LIST_HISTORY )
		{
			Trace( "FlatFile_SortFiles:  history list does not get sorted, it is an MRU list." );
		}
		else
		{
			Trace( "FlatFile_SortFiles:  sorting list %u...", iList );
			s_rgprgFiles[iList]->Sort( s_fSortByPath ? Cmp_SortByPath : Cmp_SortByName );
			Trace( "FlatFile_SortFiles:  sorted %u files.", s_rgprgFiles[iList]->Count() );
		}
	}
}

extern "C" unsigned VSAPI
FlatFile_Count( ListType iList )
{
	if( iList < 0 || iList >= LIST_MAX || !s_rgprgFiles[iList] )
		return 0;

	return s_rgprgFiles[iList]->Count();
}

/*
 * File list filtering functions.
 */

unsigned
Filter_Count()
{
	if( !s_prgMatches )
	{
		Filter_ApplyFilter( 0 );
		if( !s_prgMatches )
			return 0;
	}

	return s_prgMatches->Count();
}

FileInfo*
Filter_GetFile( unsigned index )
{
	if( !s_prgMatches )
	{
		Filter_ApplyFilter( 0 );
		if( !s_prgMatches )
			return 0;
	}

	if( index >= s_prgMatches->Count() )
		return 0;
	return *s_prgMatches->Item( index );
}

unsigned
Filter_ApplyFilter( char const* _pszFilter )
{
	bool fAnyFilterSpecified = _pszFilter && *_pszFilter;
	bool fFindFirstPrefixMatch = fAnyFilterSpecified && s_nAlwaysSubstringMatch;
	unsigned iSelect = 0;

	// Reset the matches array.

	delete s_prgMatches;
	s_prgMatches = new Rg<FileInfo*>;

	// Split the lookup text into path and file components because the list is
	// sorted by file rather than by path.

	Str sFile;
	Str sPath;
	if( _pszFilter )
	{
		char const* const _pszFile = GetFileName( _pszFilter );
		sFile.Set( _pszFile );
		if( _pszFile > _pszFilter )
		{
			sPath.Set( "*" );
			sPath.Append( _pszFilter, _pszFile - _pszFilter );
		}
		if( s_nAlwaysSubstringMatch )
		{
			sFile.Set( "*" );
			sFile.Append( _pszFile );
			sFile.Append( "*" );
			if( s_nAlwaysSubstringMatch == 2 && sPath.Length() )
			{
				Assert( sPath.Length() > 0 );
				Assert( sPath.Text()[sPath.Length() - 1] == '\\' );
				sPath.SetLength( sPath.Length() - 1 );
				sPath.Append( "*\\" );
			}
		}
	}

	Trace( "FILTERING...  (path='%s', file='%s')", sPath.Text(), sFile.Text() );

	char const* const pszFile = sFile.Text();

	// Choose which file lists to combine into a unified filtered list.

	Rg<FileInfo>* rgprg[2] = {};
	unsigned rgindex[2] = {};

	switch( s_iMode )
	{
	case MODE_SMART:
		if( fAnyFilterSpecified )
			rgprg[0] = g_fSmartIncludesWorkspace ? s_rgprgFiles[LIST_WORKSPACE] : s_rgprgFiles[LIST_PROJECT];
		rgprg[1] = s_rgprgFiles[LIST_FILESYSTEM];
		break;
	case MODE_FILESYSTEM:
		rgprg[1] = s_rgprgFiles[LIST_FILESYSTEM];
		break;
	case MODE_PROJECT:
		rgprg[0] = s_rgprgFiles[LIST_PROJECT];
		break;
	case MODE_WORKSPACE:
		rgprg[0] = s_rgprgFiles[LIST_WORKSPACE];
		break;
	case MODE_HISTORY:
		rgprg[0] = s_rgprgFiles[LIST_HISTORY];
	default:
		Panic();
		break;
	}

	Assert( FImplies( rgprg[0], rgprg[0] != s_rgprgFiles[LIST_FILESYSTEM] ) );

	unsigned const rgcount[2] = { rgprg[0] ? rgprg[0]->Count() : 0, rgprg[1] ? rgprg[1]->Count() : 0 };

	// Try to efficiently skip to the first possible match in the
	// project/workspace files.

	bool const fBeginsWithWildcard = ( *pszFile == '*' || *pszFile == '?' );
	bool const fHasWildcards = ( strchr( pszFile, '*' ) || strchr( pszFile, '?' ) );

	if( *pszFile &&						// No point skipping if no file component.
		!fBeginsWithWildcard &&			// Can't skip if first character is a wildcard.
		s_iMode != MODE_HISTORY &&		// History list is an MRU.
		!s_fSortByPath )				// Can't skip when sorted by path.
	{
		if( rgprg[0] )
			rgprg[0]->FSearch( reinterpret_cast<CompareFn>( fHasWildcards ? Cmp_NonWildPrefix : Cmp_Lookup ), pszFile, &rgindex[0] );
	}

	// Simultaneously filter and merge-sort the appropriate file list arrays,
	// to combine their results into a single filtered list.

	FileInfo* rgpentry[2];

	while( rgindex[0] < rgcount[0] || rgindex[1] < rgcount[1] )
	{
		if( rgprg[0] && rgindex[0] < rgcount[0] )
			rgpentry[0] = rgprg[0]->Item( rgindex[0] );
		else
			rgpentry[0] = 0;

		if( rgprg[1] && rgindex[1] < rgcount[1] )
			rgpentry[1] = rgprg[1]->Item( rgindex[1] );
		else
			rgpentry[1] = 0;

		Assert( rgpentry[0] || rgpentry[1] );

		FileInfo const* p = 0;
		int iSide;
		bool fCanMatchPath = true;

		if( !rgpentry[1] )
		{
			iSide = 0;
			p = rgpentry[iSide];
			rgindex[iSide]++;
		}
		else if( !rgpentry[0] )
		{
			iSide = 1;
			p = rgpentry[iSide];
			rgindex[iSide]++;

			fCanMatchPath = false;
		}
		else
		{
			// Merge-sort the two lists.

			int const n = ( s_fSortByPath ?
							Cmp_SortByPath( rgpentry[0], rgpentry[1] ) :
							Cmp_SortByName( rgpentry[0], rgpentry[1] ) );

			Trace( "%s%s -VS- %s%s", rgpentry[0]->m_sPath.Text(), rgpentry[0]->m_sFile.Text(), rgpentry[1]->m_sPath.Text(), rgpentry[1]->m_sFile.Text() );

			iSide = ( n >= 0 ) ? 1 : 0;
			p = rgpentry[iSide];

			Trace( "  >> CHOSE %s%s", p->m_sPath.Text(), p->m_sFile.Text() );

			// Increment the index from whichever side we selected.
			// Duplicates are not allowed, so if they equal then increment the
			// index from both sides.

			if( n <= 0 )
				rgindex[0]++;
			if( n >= 0 )
				rgindex[1]++;

			fCanMatchPath = ( n <= 0 );
		}

		Assert( p );

		// Test if the item matches the lookup text.

		bool const fMatchPath = ( !fCanMatchPath || ( !sPath.Length() || 0 == Cmp_WildNoCase( sPath.Text(), p->m_sPath.Text() ) ) );
		bool const fMatchFile = ( fMatchPath && ( !*pszFile || 0 == Cmp_Lookup( pszFile, p ) ) );

		if( fMatchFile )
		{
			FileInfo** pp;
			s_prgMatches->Append( &pp );
			if( !pp )
				break;
			*pp = const_cast<FileInfo*>( p );

			if( fFindFirstPrefixMatch )
			{
				Assert( pszFile );
				Assert( *pszFile == '*' );

				if( 0 == Cmp_WildNoCase( pszFile + 1, p->m_sFile.Text() ) )
				{
					iSelect = s_prgMatches->Count() - 1;
					fFindFirstPrefixMatch = false;
				}
			}
		}

		// If the file didn't match then some processing is needed to figure
		// out when to exit the loop.

		if( !fMatchFile )
		{
			// Always scan the entire LIST_FILESYSTEM list.
			if( iSide == 1 )
				continue;
			// Always scan the entire LIST_HISTORY list.
			if( s_iMode == MODE_HISTORY )
				continue;
			// If the lookup text has a path, then we have to scan the entire
			// array because it is sorted by file not by path.
			if( sPath.Length() )
				continue;
			// If the lookup text has wildcards but does not begin with a
			// wildcard then we have to scan the entire array because it is
			// sorted by file not by path.
			if( fHasWildcards && !fBeginsWithWildcard )
				continue;
			// When sorted by path we have to scan the entire array.
			if( s_fSortByPath )
				continue;
			// If the lookup text has wildcards and begins with a non-wildcard
			// character then we only have to scan while the prefix matches.
			Assert( iSide == 0 );
			if( fHasWildcards && fBeginsWithWildcard )
				if( 0 == Cmp_NonWildPrefix( pszFile, p ) )
					continue;
			// Otherwise failure to match means we're done with this side.
			rgindex[iSide] = rgcount[iSide];
			break;
		}
	}

	return iSelect;
}

